/*{{{  output.c*/
/************************************************
PROBLEM WITH TYPE PREFIX  (i.e. stmt ("%n", sp)):
-- How do we decide appropriate type for an 'sp' that has no type flag
set, but has 'AT_STMT_VAR' set.  Note that this flag may hang around
long after the stmt concerned!  (If we know it really *is* a stmt
var currently, then we can use 'lookup_type(sp)' to find it!).
************************************************/
/*{{{  #include's*/
#ifndef STDIO_H
#include <stdio.h>
#define STDIO_H
#endif

#ifdef  __STDC__
#include <stdarg.h>	/* ANSI C version */
#else
#include <varargs.h>	/* 'K & R' C version */
#endif

#include "adaptdat.h"	/* for 'proc_size', etc */
#include "expr.h"	/* for 'Expr *func_ref()' */
#include "ida.h"	/* for 'list_stack', used in 'yyerror()' */
#include "label.h"
#include "lex.h"	/* for 'in_HPF_dir', 'lineno', 'give_warnings',
			  'src_filename', 'start/end_tk', 'stmt_startp/endp'*/
#include "map.h"	/* for 'seg_extent ()' */
#include "output.h"
#include "parser.h"	/* for 'yyparse' vars 0'ed in 'yyerror()' */
#include "strmacs.h"
#include "unit.h"	/* for 'reset_normal_scope' in 'yyerror(), and
			   for 'procname_msg() */
/*}}}*/

/*{{{  globals in 'output.h'*/
bool  addto_yyerror_msg    = FALSE,
      compile              = TRUE,
      has_errors           = FALSE,
      start_of_stmt_lineno = FALSE,
      earlier_error        = FALSE,
      suppress_yyerror_msg = FALSE,
      use_current_lineno   = FALSE;

int   stmt_indent = 6,
      test_indent = 0;
/*}}}*/
/*{{{  globals for 'program output'*/
static int outcol;

#define  writechar(c)	((outcol++ <= MAX_F77_LINELEN) ?		\
			putchar((c)) :					\
			(fputs ("\n     *", stdout), putchar ((c)), outcol = 8))

#define  writestr(p,s)	for (p = (s); *p; p++) writechar (*p)
/*}}}*/

/*----------------*/
/* Error messages */
/*----------------*/
/*{{{  yyerror (char *s) -- DONE*/
yyerror (s)
/*-------*/
	char  *s;
/*----------------------------------------------------------------
	This function is called immediately by 'yyparse ()' on
finding a syntax error (before any error token is shifted or
error rule reduced).
	It tries to give a more helpful error message than just
"parse error" (the input argument!).

N.B.
---
(1)  We assume we always want to skip the rest of the stmt on
finding a syntax error, i.e. we expect the error rule to be:
                 xxx  error  '\n'
Hence on return 'end_tk' is left pointing to the EOL in 'textbuf',
so 'yylex' returns '\n' the next time it's called.  This is important
-- some error rules depend on it and will be invoked in an infinite
loop otherwise!

(2)  Every error rule must have an action containing:
		yyerrok;
This ensures that this function is *always* called when an error rule
is reduced, so 'end_tk' is always advanced to the EOL, which is important
for some errors (see last point).

(3)  It would be nice to design error rules for 'expr' that checked
for unbalanced parentheses, but this is very hard in Fortran.
----------------------------------------------------------------*/
{
	if (str_match (s, "parse error")) {
		/*{{{  give a "syntax error" message, and cleanup.*/
		if (!suppress_yyerror_msg) {
			/*--------------------------------------*/
			/* Try to give a more helpful message!	*/
			/*--------------------------------------*/
			char  *stmt_str;
			int  c;
		
			stmt_str = (in_HPF_dir ? "HPF directive" : "statement");
		
			if (start_tk == stmt_startp) {
				/*------------------------------*/
				/* error at start of stmt	*/
				/*------------------------------*/
				delimit_word_in_textbuf (c, end_tk);
				errmsg ("syntax error at start of %s, at token %q",
				         stmt_str, start_tk);
				restore_textbuf (c, end_tk);
				skip_rest_of_stmt ();
			}
			else if (end_tk > start_tk) {
				/*----------------------*/
				/* error within stmt	*/
				/*----------------------*/
				delimit_word_in_textbuf (c, end_tk);
				if (addto_yyerror_msg)
					begin_errmsg ("syntax error at or before token %q",
					               start_tk);
				else
					errmsg ("syntax error at or before token %q", start_tk);
		
				restore_textbuf (c, end_tk);
				skip_rest_of_stmt ();
			}
			else if (*start_tk == ';')
				/*------------------------------*/
				/* error at end of stmt, at ';'	*/
				/*------------------------------*/
				errmsg ("incomplete %s (terminated by \';\')", stmt_str);
		
			else
				/*------------------------------*/
				/* error at end of stmt, at '\n'*/
				/*------------------------------*/
				errmsg ("incomplete %s", stmt_str);
		}
		else
			suppress_yyerror_msg = FALSE;
		
		/*{{{  if we're in an arg-list, clear 'list_stack'*/
		/*------------------------------------------------------*/
		/* If we're in an arg-list, clear the 'list_stack'!	*/
		/*------------------------------------------------------*/
		if (list_stack) {
			Exprlist  *next_listp,  *listp;
		
			for (listp = list_stack;  listp;  listp = next_listp) {
				next_listp = listp->next;
		
				free ((char *) listp);
			}
			list_stack = NULL;
		}
		/*}}}*/
		
		/*{{{  zero 'yyparse()' vars which must be 0'ed at end of certain stmts*/
		/*--------------------------------------------------------------*/
		/* Zero 'yyparse' vars which must be 0'ed at end of certain	*/
		/* statements.							*/
		/*								*/
		/* N.B. The following 'delete_xxx()' actions are only effective	*/
		/* if the structures are unused  (i.e. nlinks == 0 for		*/
		/* 'ref_arp', nlinks == 1 for 'ref_alignp' & 'ref_distp').	*/
		/*--------------------------------------------------------------*/
		attrs = type = 0;
		if (ref_char_len) {
			delete_expr (ref_char_len, LEAVE_NLINKS);
			ref_char_len = NULL_Expr;
		}
		if (ref_arp) {
			delete_array (ref_arp, LEAVE_NLINKS);
			ref_arp = NULL_Array;
		}
		if (ref_alignp) {
			delete_align (ref_alignp, DECR_NLINKS);
						/* ... 'DECR_NLINKS' as it was created
						   with (nlinks == 1) for 'protection' */
			ref_alignp = NULL_Align;
		}
		if (ref_distp) {
			delete_distr (ref_distp, DECR_NLINKS);
						/* ... 'DECR_NLINKS' as it was created
						   with (nlinks == 1) for 'protection' */
			ref_distp = NULL_Distr;
		}
		decl_sp   = NULL_Sym;
		/*}}}*/
		
		reset_normal_scope ();
		/*}}}*/
	}
	else {
		/*--------------------------------------*/
		/* Probably "parser stack overflow"!	*/
		/*--------------------------------------*/
		errmsg (s);
		exit (1);
	}
}
/*}}}*/
/*{{{  fatal_error ()*/
fatal_error ()
/*----------*/
{
	write_err_lineno ();
	fputs ("(fatal error) cannot recover from the last error.  Goodbye!\n",
	        stderr);
	exit (1);
}
/*}}}*/
/*{{{  write_err_lineno ()*/
write_err_lineno ()
/*---------------*/
/*--------------------------------------------------------------*/
/*	Print the filename (if any) and line number of the	*/
/* location of the error that is about to be reported.  This	*/
/* information may have the following formats (depending on	*/
/* the values of 'src_filename' & 'earlier_error'):		*/
/*								*/
/*  "crap.f", line 10:						*/
/*  "crap.f", detected at line 10:				*/
/*								*/
/*  line 10:							*/
/*  detected at line 10:					*/
/*--------------------------------------------------------------*/
{
	int  err_lineno;
	
	/*------------------------------------------------------*/
	/* find 'err_lineno', the lineno to be reported in the	*/
	/* error message.					*/
	/*------------------------------------------------------*/
	if (use_current_lineno)		/* when inputting the next stmt */
		err_lineno = lineno;

	else if (start_of_stmt_lineno)
		err_lineno = stmt_lineno [0];

	else {
		/*---------------------------------*/
		/* lineno of current posn in stmt! */
		/*---------------------------------*/
		int  i;

		for (i = 0;  start_tk > stmt_endp [i];  i++)
			;
		err_lineno = stmt_lineno [i];
	}

	/*----------------------------------------------*/
	/* write the filename (if any) and lineno.	*/
	/*----------------------------------------------*/
	if (src_filename)
		fprintf (stderr, "\"%s\", ", src_filename);

	if (earlier_error)
		fprintf (stderr, "detected at line %d: ", err_lineno);
	else
		fprintf (stderr, "line %d: ", err_lineno);
}
/*}}}*/
/*{{{  procname_msg ()*/
procname_msg ()
/*-----------*/
/*--------------------------------------------------------------*/
/* Print the name of the next subprogram to be 'compiled', in	*/
/* a format like:						*/
/*								*/
/*	> module M1:  subroutine S1:  internal function F1:	*/
/*								*/
/* where 'M1', 'S1' and 'F1' are the subprogram names.		*/
/*--------------------------------------------------------------*/
{
	Unit  *unitp,
	      *last_unitp = 0;	/* the last subprogram name written-out */

	for (last_unitp = 0;
	     last_unitp != this_unitp;
	     last_unitp = unitp) {

		/*----------------------------------------------*/
		/* Find 'unitp' such that 'unitp->prev' is	*/
		/* 'last_unitp', the 'unit record' of the last	*/
		/* subprogram whose name was written out...	*/
		/*----------------------------------------------*/
		for (unitp = this_unitp;
		     unitp->prev != last_unitp  &&  !(unitp->attrs & INTERFACE_LEVEL);
		     unitp = unitp->prev)
			;
		/*----------------------------------------------*/
		/* Don't output the name of any interface body!	*/
		/*----------------------------------------------*/
		if (unitp->attrs & INTERFACE_LEVEL)
			return;

		/*----------------------------------------------*/
		/* Now write out the type and name of the	*/
		/* subprogram recorded in 'unitp'.		*/
		/*----------------------------------------------*/
		fputs ("  ", stderr);
		if (unitp->attrs & INTERNAL_LEVEL)
			fputs ("internal ", stderr);
			
		switch (unitp->attrs & UNIT_PROC_TYPE) {
		case  SUBROUTINE_UNIT :	fputs ("subroutine", stderr);	break;
		case  FUNCTION_UNIT :	fputs ("function", stderr);	break;
		case  BLOCK_DATA_UNIT :	fputs ("block data", stderr);	break;
		case  PROGRAM_UNIT :	fputs ("main", stderr);		break;
		case  MODULE_UNIT :	fputs ("module", stderr);	break;
		}
		if (unitp->sp)
			fprintf (stderr, " %s", unitp->sp->name);
		putc (':', stderr);
	}
	putc ('\n', stderr);
}
/*}}}*/

/*--------------------------------------------------------------------*/
/* Program output  (only for 'ADAPT'; empty unless 'V_ADAPT' defined) */
/*--------------------------------------------------------------------*/
/*{{{  write_stmt_label ()*/
write_stmt_label ()
/*---------------*/
{
#ifdef  V_ADAPT
	register int  i;

	if (has_errors)
		return;		/* don't generate output */

	if (stmt_newlabel) {
		printf ("%5d", stmt_newlabel);
		stmt_newlabel = 0;
		for (i = 6; i <= stmt_indent; i++)
			putchar (' '); 
	}
	else {
		for (i = 1; i <= stmt_indent; i++)
			putchar (' ');
	}
	outcol = stmt_indent + 1;
#endif
}
/*}}}*/

/*------------------------------------------------------*/
/* Expression output  (the last 2 only used in 'ida.c') */
/*------------------------------------------------------*/
/*{{{  stmt_write_expr (ep, context, parent_op, branch) -- DONE VEC*/
stmt_write_expr (ep, context, parent_op, branch)
/*--------------------------------------------*/
	Expr  *ep;
	int    context;		/* F90_ / ELEMENTALISED_ / ARG_ /
				   ARG_PLUS_MAPVEC_EXPR.  This affects
				   the o/p format of array primaries */
	int32  parent_op;
	int32  branch;		/* 0, ND_LEFT_BRANCH or ND_RIGHT_BRANCH */
/*--------------------------------------------------------------*/
/*	Called by 'write_assignment_stmt ()' to output the	*/
/* elemental form of a variable or an expression in a		*/
/* transformed array assignment (which is now transformed to	*/
/* an elemental assignment enclosed in DO-loops and/or IF	*/
/* constructs).  It's also called recursively to output a	*/
/* subexpression of that expression.  In the former case	*/
/* 'parent_op' and 'branch' are both 0.				*/
/*--------------------------------------------------------------*/
{
	register  char *p;	/* used by 'writestr()' to traverse strings */

	if (!ep)	/* NULL pointer, eg, to left subtree of a unary op */
		return;

/**************
-- PROBLEM WITH WRITING SUBSCR DOUBLETS/TRIPLETS -- DOESN'T REFER TO
sectp->l [dim], ETC.
-- ENSURE ARRAY SUBSCRIPTS EVERYWHERE REFER TO ENTRIES IN 'sectp'!
*************************/
	switch (ep->token) {
	/*--------------------------------------*/
	/* '*ep' is a 'leaf' (i.e. operand)...	*/
	/*--------------------------------------*/
	/*{{{  case  ND_CHAR / COMPLEX / DP / REAL_LIT_CONST : -- OK*/
	case  ND_CHAR_LIT_CONST :
	case  ND_COMPLEX_LIT_CONST :
	case  ND_DP_LIT_CONST :
	case  ND_REAL_LIT_CONST :
		/*------------------------------------------------------*/
		/* Print string and free it (we assume that it's only	*/
		/* wanted once!)					*/
		/*							*/
		/* N.B.  We *don't* free the string associated with a	*/
		/* 'ND_LOGICAL_LIT_CONST', as that points to the keytab!*/
		/*------------------------------------------------------*/
		writestr (p, ep->val.strval);
		free (ep->val.strval);
		
		if (ep->left) {
			/*----------------------------------------------*/
			/* 'token' should be 'ND_CHAR_LIT_CONST', and	*/
			/* it has a substring range (i.e. 'ep->left'	*/
			/* should be a 'ND_SUBSCR_DOUBLET' node').	*/
			/*----------------------------------------------*/
			writechar ('(');
			stmt_write_expr (ep->left, F90_EXPR, 0L, 0L);
			writechar (')');
		}
		
		return;
	
	/*}}}*/
	/*{{{  case  ND_LOGICAL_LIT_CONST :  ND_INTERN_OBJECT : -- OK*/
	case  ND_LOGICAL_LIT_CONST :
	case  ND_INTERN_OBJECT :
		writestr (p, ep->val.strval);
		return;
	
	/*}}}*/
	/*{{{  case  ND_INT_LIT_CONST : -- OK*/
	case  ND_INT_LIT_CONST :
		cont_stmt ("%d", ep->val.intval);
		return;
	
	/*}}}*/
	/*{{{  case  ND_ALT_RETURN_LABEL : -- OK*/
	case  ND_ALT_RETURN_LABEL :
		cont_stmt ("*%d", ep->val.intval);
		return;
	/*}}}*/
	/*{{{  case  ND_TEMP_SCALAR : -- OK*/
	case  ND_TEMP_SCALAR :
		cont_stmt ("%pV%d",  ep->type,  ep->val.allocp->num);
		return;
	/*}}}*/
	/*{{{  case  ND_SIMPLE_SCALAR :  PROC_ACTUAL_ARG  -- OK*/
	case  ND_SIMPLE_SCALAR :
	case  ND_PROC_ACTUAL_ARG :
		/*--------------------------------------*/
		/* Print  name [(substring range)]	*/
		/*--------------------------------------*/
		cont_stmt ("%n", ep->val.sp);
		
		if (ep->left) {
			/*----------------------------------------------*/
			/* 'token' should be 'ND_SIMPLE_SCALAR', and it	*/
			/* has a substring range  (i.e. 'ep->left'	*/
			/* should be a 'ND_SUBSCR_DOUBLET' node).	*/
			/*----------------------------------------------*/
			writechar ('(');
			stmt_write_expr (ep->left, F90_EXPR, 0L, 0L);
			writechar (')');
		}
		return;
	
	/*}}}*/
	/*{{{  case  ND_WHOLE_ARRAY :  ARRAY_SECT :  TEMP_ARRAY --OK*/
	case  ND_WHOLE_ARRAY :
	case  ND_ARRAY_SECT :
	case  ND_TEMP_ARRAY :
	{
		/*------------------------------------------------------*/
		/* Output a reference to an array or temp array section	*/
		/* or element.  'ep->mapp' and 'ep->sectp' are defined.	*/
		/*------------------------------------------------------*/
		bool    temparray = (ep->token == ND_TEMP_ARRAY);
		int     array_rank,  dim,  n_sect_dim;
		Expr  **lp,  **np,  **sp;
		Map    *mapp  = ep->mapp;		/* != NULL */
		Sect   *sectp = ep->sectp;		/* != NULL */
	
		switch (context) {
		case  ARG_EXPR :
		case  ARG_PLUS_MAPVEC_EXPR :
			/*{{{  an actual arg;  if not an array element, o/p name only*/
			/*--------------------------------------------------------------*/
			/* Array reference is an actual argument.			*/
			/*	If it's an array section or whole array, output just	*/
			/* its name;  if it's an array *element*, output its subscript	*/
			/* list as well.						*/
			/*--------------------------------------------------------------*/
			if (temparray) {
				/*------------------------------------------------------*/
				/* Write reference to start of temp array  (e.g.:	*/
				/*	"T_TMP(K_T2)"					*/
				/*------------------------------------------------------*/
				cont_stmt ("%pTMP(K_%c%d)",
				             ep->type,
				           ((ep->val.allocp->alloc_type == ALLOC_ARRAY) ? 'T' : 'M'),
				             ep->val.allocp->num);
			
				if (context == ARG_PLUS_MAPVEC_EXPR) {
					writestr (p, ", ");
					stmt_write_expr (sectp->mapvec_ep, ARG_EXPR, 0L, 0L);
				}
			}
			else if ( ! (ep->type & EX_SCALAR)) {
				/*------------------------------------------------------*/
				/* A whole array or array section.  Just write the	*/
				/* array name.						*/
				/*------------------------------------------------------*/
				cont_stmt ("%n", ep->val.sp);
			
				if (context == ARG_PLUS_MAPVEC_EXPR) {
					writestr (p, ", ");
					stmt_write_expr (sectp->mapvec_ep, ARG_EXPR, 0L, 0L);
				}
			}
			else {
				/*------------------------------------------------------*/
				/* An array *element*.  Write its subscript list (and	*/
				/* *don't* write a 'mapvec').				*/
				/*------------------------------------------------------*/
				cont_stmt ("%n", ep->val.sp);
				
				array_rank = ((ep->mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
				
				for (dim = 0,  lp = ep->sectp->l;    dim < array_rank;    ++dim,  ++lp) {
					/*------------------------------------------------------*/
					/* Write subscript expression.				*/
					/* N.B. We expect *all* subscripts to be scalar.	*/
					/*------------------------------------------------------*/
					writechar (dim ? ',' : '(');
					stmt_write_expr (*lp, F90_EXPR, 0L, 0L);
				}
				writechar (')');
			}
			return;
			/*}}}*/
	
		case  F90_EXPR :	/* assume that the array is collapsed */
			/*{{{  write array ref with its subscript list (if any) in F90 style*/
			/*--------------------------------------------------------------*/
			/* Write out the array reference in F90 style, viz:		*/
			/*	"array-name ([subscript list])"				*/
			/*								*/
			/* N.B. This assumes that the array is *collapsed* on the	*/
			/* node(s) executing this code.					*/
			/*--------------------------------------------------------------*/
			if (temparray) {		/* assume it's a temp var, *not* a mapvec */
				/*--------------------------------------------------------
				Write reference to the whole of temp array, e.g.:
					"T_TMP(K_T2 : K_T2+len-1)"
			
				N.B. If this in an array expr or assignment, the shape may
				well not conform.  Hence it's only safe to use in isolation
				in an I/O list, or as a vector subscript therein.
				--------------------------------------------------------*/
				Expr  *len_m1 = expr (seg_totlen (NULL_Sym, mapp), ND_MINUS_OP, int_lit_1);
			
				cont_stmt ("%pTMP(K_T%d : K_T%d+(%e))",
				             ep->type,  
				             ep->val.allocp->num,  ep->val.allocp->num,  len_m1);
			
				delete_expr (len_m1, LEAVE_NLINKS);
			}
			else if (ep->token == ND_WHOLE_ARRAY)
				cont_stmt ("%n", ep->val.sp);
			
			else {	/*----------------------------------------------*/
				/* A 'ND_ARRAY_SECT'.				*/
				/* N.B. We assume it's *collapsed* on the node.	*/
				/*----------------------------------------------*/
				cont_stmt ("%n", ep->val.sp);
			
				lp = sectp->l;
				np = sectp->n;
				sp = sectp->s;
				array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
			
				for (dim = 0;    dim < array_rank;    ++dim,  ++lp,  ++np,  ++sp) {
					writechar (dim ? ',' : '(');
					/*{{{  write subscript expression*/
					/*------------------------------*/
					/* Write subscript expression	*/
					/*------------------------------*/
					if (*np == int_lit_m1  ||  *np == int_lit_m2)
						/*------------------------------------------------------*/
						/* A single element or vector subscript in (collapsed)	*/
						/* dimension 'dim'.  Simply write 'l'.			*/
						/*------------------------------------------------------*/
						stmt_write_expr (*lp, F90_EXPR, 0L, 0L);
					
					else {
						/*------------------------------------------------------*/
						/* A regular section in a (collapsed) dimension 'dim'.	*/
						/* Write:						*/
						/*	"l : u [: s]"					*/
						/* where:						*/
						/*	l = *lp						*/
						/*	s = *sp						*/
						/*	u = l + (n-1)*s					*/
						/*------------------------------------------------------*/
						Expr  *u;
					
						u = expr (*lp, ND_PLUS_OP, expr (*np, ND_TIMES_OP, *sp));
					
						cont_stmt ("%e:%e", *lp, u);
						if (*sp != int_lit_1)
							cont_stmt (":%e", *sp);
					
						delete_expr (u, LEAVE_NLINKS);
					}
					/*}}}*/
				}
				writechar (')');
			}
			return;
			/*}}}*/
	
		default :
			/*{{{  in assignment stmt; o/p ref with subscripts 'elementalised'*/
			/*--------------------------------------------------------------*/
			/* Array reference is in an 'elementalised' assignment stmt.	*/
			/* Hence output the reference with its subscripts		*/
			/* 'elementalised'.						*/
			/*--------------------------------------------------------------*/
			if (temparray) {
				/*{{{  write start of temp array ref (e.g. "T_TMP (K_T2+")*/
				/*------------------------------------------------------*/
				/* Write start of temp array ref (e.g. "T_TMP (K_T2+")	*/
				/*------------------------------------------------------*/
				cont_stmt ("%pTMP (K_T%d+",  ep->type,  ep->val.allocp->num);
				/*}}}*/
			}
			else
				cont_stmt ("%n (", ep->val.sp);
			
			lp = sectp->l;
			np = sectp->n;
			sp = sectp->s;
			array_rank = ((mapp->attrs & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
			n_sect_dim = context;		/* the number of the first section dim
							   (0 unless 'ep' is a vector subscript) */
			
			for (dim = 0;    dim < array_rank;    ++dim,  ++lp,  ++np,  ++sp) {
				if (dim) {
					/*{{{  write "+ locn*(" for ND_TEMP_ARRAY, "," otherwise*/
					if (temparray) {
						/*------------------------------------------------------*/
						/* Write:						*/
						/*	"+ (locn)*("					*/
						/* where 'locn' is the size of the local segment in	*/
						/* the previous dimension, i.e. (dim-1).		*/
						/*------------------------------------------------------*/
						Expr  *locn = seg_extent (NULL_Sym, mapp, (dim - 1));
					
						cont_stmt ("+(%e)*(", locn);
					
						delete_expr (locn, LEAVE_NLINKS);
					}
					else
						writechar (',');
					/*}}}*/
				}
				/*{{{  write subscript expression*/
				/*------------------------------*/
				/* Write subscript expression	*/
				/*------------------------------*/
				if (*np == int_lit_m1)
					/*------------------------------------------------------*/
					/* A single element in dimension 'dim' (collapsed or	*/
					/* distributed).  Simply write:				*/
					/*	l						*/
					/* (which is set to 'loc_l' if it's a distributed dim).	*/
					/*------------------------------------------------------*/
					stmt_write_expr (*lp, F90_EXPR, 0L, 0L);
				
				else if (*np == int_lit_m2) {
					/*------------------------------------------------------*/
					/* A vector subscript in (collapsed) dimension 'dim'.	*/
					/*------------------------------------------------------*/
					stmt_write_expr (*lp, n_sect_dim, 0L, 0L);
					++n_sect_dim;		/* the *next* section dim */
				}
				else {
					/*------------------------------------------------------*/
					/* A regular section in dimension 'dim'	(collapsed or	*/
					/* distributed).  Write:				*/
					/*	"l + do_var * s"				*/
					/* where:						*/
					/*	do_var = ("K_%d", n_sect_dim)			*/
					/*							*/
					/*	'l' and 's' already store the appropriate	*/
					/* values.  E.g., for a distributed dim they store	*/
					/* 'loc_l' & 'loc_s'.  Likewise, for a collapsed dim on	*/
					/* the RHS of an assignment, if the corresponding dim	*/
					/* of the assignment variable is *distributed* then 'l'	*/
					/* stores (l + eta * s) as required.			*/
					/*------------------------------------------------------*/
					char   strbuf [5];
					Expr  *do_var,  *tmp_ep;
				
					sprintf (strbuf, "K_%d", n_sect_dim);		/* e.g. "K_0" */
					do_var = name_primary (strbuf, (EX_INTEGER | EX_SCALAR));
				
					tmp_ep = expr (*lp, ND_PLUS_OP, expr (do_var, ND_TIMES_OP, *sp));
				
					stmt_write_expr (tmp_ep, F90_EXPR, 0L, 0L);
				
					delete_expr (tmp_ep, LEAVE_NLINKS);
				
					++n_sect_dim;		/* the *next* section dim */
				}
				/*}}}*/
			}
			/*{{{  o/p ")" 'array_rank' times for ND_TEMP_ARRAY, once otherwise*/
			/*-----------------------------------------------------------------*/
			/* o/p ")" 'array_rank' times for ND_TEMP_ARRAY, or once otherwise */
			/*-----------------------------------------------------------------*/
			if (temparray) {
				for (dim = array_rank;   dim;   --dim)
					writechar (')');
			}
			else
				writechar (')');
			/*}}}*/
			return;
			
			
			
			/*}}}*/
		}
	}
	/*}}}*/
	/*{{{  case  ND_FUNC :  INTERN_FUNC :  SUBR :  STRUCT_CONSTR : -- OK*/
	case  ND_FUNC :
	case  ND_INTERN_FUNC :
	case  ND_SUBR :
	case  ND_STRUCT_CONSTR :
	{
		/*------------------------------------------------------*/
		/* Print  "name  ([subscript/arg/comp list])"		*/
		/*------------------------------------------------------*/
		bool   first_arg;
		int    arg_context;
		Expr  *arg_ep;
	
		if (ep->token == ND_INTERN_FUNC)
			writestr (p, ep->val.strval);
		else
			cont_stmt ("%n", ep->val.sp);
		writechar ('(');
	
		arg_context = ARG_EXPR;
		if ((ep->token == ND_INTERN_FUNC) || (ep->token == ND_STRUCT_CONSTR))
			arg_context = F90_EXPR;
	
		else if ((context >= 0)  &&  (ep->token == ND_FUNC)) {
			/*----------------------------------------------*/
			/* In an assignment expr.  If this is an	*/
			/* elemental function, 'elementalise' its args.	*/
			/*----------------------------------------------*/
			Symbol  *sp = ep->val.sp;
	
			if ((sp->attrs & AT_INFO)  &&
			    (sp->info.sp->attrs & AT_INTRINSIC)  &&
			    (sp->info.sp->type & TY_ELEMENTAL))
				arg_context = context;	/* i.e. 'elementalise' args */
		}
	
		first_arg = TRUE;
		arg_ep = ep;
		while (arg_ep = arg_ep->right) {	/* 'arg_ep' = a 'ND_EXPRLIST' node */
			if (first_arg)
				first_arg = FALSE;
			else
				writechar (',');
			/*---------------------------*/
			/* write the next actual arg */
			/*---------------------------*/
			stmt_write_expr (arg_ep->left, arg_context, 0L, 0L);
		}
		writechar (')');
		return;
	}
	/*}}}*/
	/*{{{  case  ND_SUBSCR_DOUBLET / TRIPLET -- OK*/
	/*--------------------------------------------------------------*/
	/* This should only be called in the context of a 'CASE' stmt	*/
	/*--------------------------------------------------------------*/
	case  ND_SUBSCR_DOUBLET :
		/*--------------------------------------*/
		/*	ep = ND_SUBSCR_DOUBLET		*/
		/*	        /          \		*/
		/*            [l]          [u]		*/
		/*--------------------------------------*/
		if (ep->left)
			stmt_write_expr (ep->left, F90_EXPR, 0L, 0L);
		writechar (':');
		if (ep->right)
			stmt_write_expr (ep->right, F90_EXPR, 0L, 0L);
		return;
	
	case  ND_SUBSCR_TRIPLET :
		/*--------------------------------------*/
		/*	    ep = ND_SUBSRC_TRIPLET	*/
		/*	            /          \	*/
		/*	[ ND_SUBSCR_DOUBLET ]   s	*/
		/*	[    /          \   ]		*/
		/*	[   l            u  ]		*/
		/*--------------------------------------*/
		if (ep->left) {
			/*-------------------------------------------------*/
			/* 'ep->left' should be a 'ND_SUBSCR_DOUBLET' node */
			/*-------------------------------------------------*/
			if (ep->left->left)
				stmt_write_expr (ep->left->left, F90_EXPR, 0L, 0L);
								/* [l] */
			writechar (':');			/* : */
			if (ep->left->right)
				stmt_write_expr (ep->left->right, F90_EXPR, 0L, 0L);
								/* [u] */
		}
		else
			writechar (':');			/* : */
		
		/*------------------------------------------------------*/
		/* 'ep->right' must be defined, and gives the ": s"	*/
		/* part of the subscript triplet.			*/
		/*------------------------------------------------------*/
		writechar (':');
		stmt_write_expr (ep->right, F90_EXPR, 0L, 0L);	/* : s */
		
		return;
	/*}}}*/
	/*{{{  case  ND_EXPRLIST : -- OK*/
	case  ND_EXPRLIST :
	
		if (ep->val.sp)
			/*----------------------------------------------*/
			/* a keyword argument: "keyword = ".		*/
			/*						*/
			/* N.B. We assume this will only be used for	*/
			/* F90 keywords, e.g. in arglist of I/O stmt or	*/
			/* intrinsic proc ref, *not* for an external	*/
			/* proc ref, since the keywords aren't		*/
			/* 'translated'.				*/
			/*----------------------------------------------*/
			cont_stmt ("%s = ", ep->val.sp->name);
	
		stmt_write_expr (ep->left, context, 0L, 0L);
		if (ep->right) {
			writechar (',');
			writechar (' ');
			stmt_write_expr (ep->right, context, 0L, 0L);
		}
		return;
	
	/*}}}*/
	/*{{{  case  ND_STAR : -- OK*/
	case  ND_STAR :
		writechar ('*');
		return;
	
	/*}}}*/
	/*{{{  case  ND_ERROR : -- OK*/
	case  ND_ERROR :
		writestr (p, "[error]");
		return;
	
	/*}}}*/

	/*------------------------------*/
	/* '*ep' is an operator node	*/
	/*------------------------------*/
	default :
		/*{{{  print expression tree '*ep' -- ok*/
		{
			int32  this_op        = ep->token;
			int32  prec_this_op   = this_op   & ND_PREC_MASK;
			int32  prec_parent_op = parent_op & ND_PREC_MASK;
			bool   need_brackets  = parent_op  &&
				( (prec_this_op < prec_parent_op) ||
				  (( prec_this_op == prec_parent_op) && !(this_op & branch))
				);
		
			if (need_brackets)  writechar ('(');
			stmt_write_expr (ep->left, context, this_op, ND_LEFT_BRANCH);
		
			switch (this_op) {
			case ND_DEFINED_OP :	cont_stmt ("%n", ep->val.sp);	break;
			case ND_EQV_OP :	writestr (p, ".EQV.");		break;
			case ND_NEQV_OP :	writestr (p, ".NEQV.");		break;
			case ND_OR_OP :		writestr (p, ".OR.");		break;
			case ND_AND_OP :	writestr (p, ".AND.");		break;
			case ND_NOT_OP :	writestr (p, ".NOT.");		break;
			case ND_REL_OP :	writestr (p, ep->val.strval);	break;
			case ND_CONCAT_OP :	writestr (p, "//");		break;
			case ND_PLUS_OP :	writechar ('+');		break;
			case ND_U_PLUS_OP :	writechar ('+');		break;
			case ND_MINUS_OP :	writechar ('-');		break;
			case ND_U_MINUS_OP :	writechar ('-');		break;
			case ND_TIMES_OP :	writechar ('*');		break;
			case ND_DIV_OP :	writechar ('/');		break;
			case ND_POWER_OP :	writestr (p, "**");		break;
			}
		
			stmt_write_expr (ep->right, context, this_op, ND_RIGHT_BRANCH);
			if (need_brackets)  writechar (')');
		
			return;
		}
		/*}}}*/
	}
}
/*}}}*/

/*{{{  write_expr (stream, ep, parent_op, branch)     **** UP TO DATE *****/
write_expr (stream, ep, parent_op, branch)
/*--------------------------------------*/
	FILE  *stream;
	Expr  *ep;
	int32  parent_op;
	int32  branch;		/* 0, ND_LEFT_BRANCH or ND_RIGHT_BRANCH */
/*--------------------------------------------------------------*/
/*	Called from 'errmsg ()', 'testmsg ()','write_ida_rec ()'*/
/* and elsewhere to output an expression to a 'stream', and	*/
/* also called recursively to output a subexpression of the	*/
/* expression.  In the former case, 'parent_op' & 'branch' == 0.*/
/*								*/
/* N.B.:  This is similar to 'stmt_write_expr()' except that	*/
/* it uses normal output routines instead of 'writechar()',	*/
/* 'writestr ()', 'cont_stmt()', etc, which provide the		*/
/* indentation and continuation lines required for Fortran o/p.	*/
/*--------------------------------------------------------------*/
{
	if (!ep)	/* NULL pointer, eg, to left subtree of a unary op */
		return;
		
	switch (ep->token) {
	/*----------------------------------------------*/
	/* if '*ep' is a 'leaf' (i.e. operand)...	*/
	/*----------------------------------------------*/
	/*{{{  case  ND_CHAR / COMPLEX / DP / REAL_LIT_CONST :*/
	case  ND_CHAR_LIT_CONST :
	case  ND_COMPLEX_LIT_CONST :
	case  ND_DP_LIT_CONST :
	case  ND_REAL_LIT_CONST :
		/*------------------------------------------------------*/
		/* Print string and free it (we assume that it's only	*/
		/* wanted once!)					*/
		/*							*/
		/* N.B.  We *don't* free the string associated with a	*/
		/* 'ND_LOGICAL_LIT_CONST', as that points to the keytab!*/
		/*------------------------------------------------------*/
		fputs (ep->val.strval, stream);
		free (ep->val.strval);
		
		if (ep->left) {
			/*----------------------------------------------*/
			/* 'token' should be 'ND_CHAR_LIT_CONST', and	*/
			/* it has a substring range (i.e. 'ep->left'	*/
			/* should be a 'ND_SUBSCR_DOUBLET' node').	*/
			/*----------------------------------------------*/
			putc ('(', stream);
			write_expr (stream, ep->left, 0L, 0L);
			putc (')', stream);
		}
		
		return;
	
	/*}}}*/
	/*{{{  case  ND_LOGICAL_LIT_CONST :  ND_INTERN_OBJECT :*/
	case  ND_LOGICAL_LIT_CONST :
	case  ND_INTERN_OBJECT :
		fputs (ep->val.strval, stream);
		return;
	
	/*}}}*/
	/*{{{  case  ND_INT_LIT_CONST :*/
	case  ND_INT_LIT_CONST :
		fprintf (stream, "%d", ep->val.intval);
		return;
	
	/*}}}*/
	/*{{{  case  ND_TEMP_SCALAR :*/
	case  ND_TEMP_SCALAR :
		fprintf (stream, "%c_V%d", intern_id_prefix (ep->type),
		                           ep->val.allocp->num);
		return;
	/*}}}*/
	/*{{{  case  ND_TEMP_ARRAY :*/
	case  ND_TEMP_ARRAY :
		fprintf (stream, "%c_TMP(K_%c%d)",
		                intern_id_prefix (ep->type),
		              ((ep->val.allocp->alloc_type == ALLOC_ARRAY) ? 'T' : 'M'),
		                ep->val.allocp->num);
		return;
	
	/*}}}*/
	/*{{{  case  ND_ALT_RETURN_LABEL :*/
	case  ND_ALT_RETURN_LABEL :
		fprintf (stream, "*%d", ep->val.intval);
		return;
	
	/*}}}*/
	/*{{{  case  ND_SIMPLE_SCALAR :  ND_WHOLE_ARRAY :  ND_PROC_ACTUAL_ARG*/
	case  ND_SIMPLE_SCALAR :
	case  ND_WHOLE_ARRAY :
	case  ND_PROC_ACTUAL_ARG :
		/*--------------------------------------*/
		/* Print  name [(substring range)]	*/
		/*--------------------------------------*/
		fputs (ep->val.sp->name, stream);
		
		if (ep->left) {
			/*----------------------------------------------*/
			/* 'token' should be 'ND_SIMPLE_SCALAR', and it	*/
			/* has a substring range  (i.e. 'ep->left'	*/
			/* should be a 'ND_SUBSCR_DOUBLET' node').	*/
			/*----------------------------------------------*/
			putc ('(', stream);
			write_expr (stream, ep->left, 0L, 0L);
			putc (')', stream);
		}
		
		return;
	
	/*}}}*/
	/*{{{  case  ND_ARRAY_SECT / _FUNC / _INTERN_FUNC / _SUBR / _STRUCT_CONSTR :*/
	case  ND_ARRAY_SECT :
	case  ND_FUNC :
	case  ND_INTERN_FUNC :
	case  ND_SUBR :
	case  ND_STRUCT_CONSTR :
		/*---------------------------------------------------------------*/
		/* Print  name  ([subscript/arg/comp list])  [(substring range)] */
		/*---------------------------------------------------------------*/
	{
		bool  first_arg = TRUE;
		Expr  *arg_ep;
		
		fprintf (stream, "%s(", ((ep->token == ND_INTERN_FUNC) ?
		                          ep->val.strval : ep->val.sp->name) );
	
	
		arg_ep = ep;
		while (arg_ep = arg_ep->right) {	/* 'arg_ep' = a 'ND_EXPRLIST' node */
			if (first_arg)
				first_arg = FALSE;
			else
				putc (',', stream);
	
			write_expr (stream, arg_ep->left, 0L, 0L);  /* an actual arg
								       or subscript */
		}
		putc (')', stream);
	
		if (ep->left) {
			/*----------------------------------------------*/
			/* 'token' should be 'ND_ARRAY_SECT', and it	*/
			/* has a substring range (i.e. 'ep->left'	*/
			/* should be a 'ND_SUBSCR_DOUBLET' node').	*/
			/*----------------------------------------------*/
			putc ('(', stream);
			write_expr (stream, ep->left, 0L, 0L);
			putc (')', stream);
		}
		return;
	}
	/*}}}*/
	/*{{{  case  ND_SUBSCR_DOUBLET / TRIPLET*/
	case  ND_SUBSCR_DOUBLET :
		/*--------------------------------------*/
		/*	ep = ND_SUBSCR_DOUBLET		*/
		/*	        /          \		*/
		/*            [l]          [u]		*/
		/*--------------------------------------*/
		if (ep->left)
			write_expr (stream, ep->left, 0L, 0L);
		putc (':', stream);
		if (ep->right)
			write_expr (stream, ep->right, 0L, 0L);
		return;
	
	case  ND_SUBSCR_TRIPLET :
		/*--------------------------------------*/
		/*	    ep = ND_SUBSRC_TRIPLET	*/
		/*	            /          \	*/
		/*	[ ND_SUBSCR_DOUBLET ]   s	*/
		/*	[    /          \   ]		*/
		/*	[   l            u  ]		*/
		/*--------------------------------------*/
		if (ep->left) {
			/*-------------------------------------------------*/
			/* 'ep->left' should be a 'ND_SUBSCR_DOUBLET' node */
			/*-------------------------------------------------*/
			if (ep->left->left)
				write_expr (stream, ep->left->left, 0L, 0L);	/* [l] */
			putc (':', stream);					/* : */
			if (ep->left->right)
				write_expr (stream, ep->left->right, 0L, 0L);	/* [u] */
		}
		else
			putc (':', stream);			/* : */
		
		/*------------------------------------------------------*/
		/* 'ep->right' must be defined, and gives the ": s"	*/
		/* part of the subscript triplet.			*/
		/*------------------------------------------------------*/
		putc (':', stream);
		write_expr (stream, ep->right, 0L, 0L);		/* : s */
		
		return;
	/*}}}*/
	/*{{{  case  ND_EXPRLIST :*/
	case  ND_EXPRLIST :
		/*------------------------------------------------------*/
		/* ND_EXPRLIST.						*/
		/* (Incidentally, we don't expect this function to be	*/
		/* used for a ND_EXPRLIST node!)			*/
		/*------------------------------------------------------*/
		write_expr (stream, ep->left, 0L, 0L);
		return;
	
	/*}}}*/
	/*{{{  case  ND_STAR :*/
	case  ND_STAR :
		putc ('*', stream);
		return;
	
	/*}}}*/
	/*{{{  case  ND_ERROR :*/
	case  ND_ERROR :
		fputs ("[error]", stream);
		return;
	
	/*}}}*/

	/*------------------------------*/
	/* '*ep' is an operator node	*/
	/*------------------------------*/
	default :
		/*{{{  print expression tree '*ep'*/
		{
			int32  this_op        = ep->token;
			int32  prec_this_op   = this_op   & ND_PREC_MASK;
			int32  prec_parent_op = parent_op & ND_PREC_MASK;
			bool   need_brackets  = parent_op  &&
				( (prec_this_op < prec_parent_op) ||
				  (( prec_this_op == prec_parent_op) && !(this_op & branch))
				);
		
			if (need_brackets)  putc ('(', stream);
			write_expr (stream, ep->left, this_op, ND_LEFT_BRANCH);
		
			switch (this_op) {
			case ND_DEFINED_OP :	fputs (ep->val.sp->name, stream); break;
			case ND_EQV_OP :	fputs (".EQV.", stream);	break;
			case ND_NEQV_OP :	fputs (".NEQV.", stream);	break;
			case ND_OR_OP :		fputs (".OR.", stream);		break;
			case ND_AND_OP :	fputs (".AND.", stream);	break;
			case ND_NOT_OP :	fputs (".NOT.", stream);	break;
			case ND_REL_OP :	fputs (ep->val.strval, stream);	break;
			case ND_CONCAT_OP :	fputs ("//", stream);		break;
			case ND_PLUS_OP :	putc ('+', stream);		break;
			case ND_U_PLUS_OP :	putc ('+', stream);		break;
			case ND_MINUS_OP :	putc ('-', stream);		break;
			case ND_U_MINUS_OP :	putc ('-', stream);		break;
			case ND_TIMES_OP :	putc ('*', stream);		break;
			case ND_DIV_OP :	putc ('/', stream);		break;
			case ND_POWER_OP :	fputs ("**", stream);		break;
			}
		
			write_expr (stream, ep->right, this_op, ND_RIGHT_BRANCH);
			if (need_brackets)  putc (')', stream);
		
			return;
		}
		/*}}}*/
	}
}
/*}}}*/

/*{{{  char  *write_expr_to_str (p, ep, parent_op, branch) *** out of date!! ****/
char  *write_expr_to_str (p, ep, parent_op, branch)
/*-----------------------------------------------*/
	char  *p;
	Expr  *ep;
	int32  parent_op;
	int32  branch;		/* 0, ND_LEFT_BRANCH or ND_RIGHT_BRANCH */
/*--------------------------------------------------------------*/
/*	Writes expr '*ep' to the string 'p' (which must be long	*/
/* enough!)  The string was previously allocated by the caller,	*/
/* using 'expr_len()' to determine the required string length.	*/
/* The function returns a pointer to the character after the	*/
/* last one written.						*/
/*	This is called by 'install_ida_proc_ref()' to initialise*/
/* the 'strval' member of an 'Idref' record.			*/
/*--------------------------------------------------------------*/
{
	if (!ep)	/* NULL pointer, eg, to left subtree of a unary op */
		return  p;

	switch (ep->token) {
	/*--------------------------------------*/
	/* '*ep' is a 'leaf' (i.e. operand):	*/
	/*--------------------------------------*/
	case  ND_SIMPLE_SCALAR :
	case  ND_WHOLE_ARRAY :
	case  ND_PROC_ACTUAL_ARG :
		strcpy (p, ep->val.sp->name);
		for ( ; *p; p++);	break;
		
	case  ND_ARRAY_SECT :
	case  ND_FUNC :
	case  ND_INTERN_FUNC :
	case  ND_SUBR :	/** Expect 'func' to be scalar, & all keyword args,
			    etc, to be resolved, only simple args left **/
		/*{{{  print name (arg list)*/
		{
			bool  not_first_arg = FALSE;
			
			strcpy (p, ((ep->token == ND_INTERN_FUNC) ?
			             ep->val.strval : ep->val.sp->name) );
			for ( ; *p; p++);
			*p++ = '(';
			
			while (ep = ep->right) {	/* ep = a 'ND_EXPRLIST' node */
				if (not_first_arg)
					*p++ = ',';
				else
					not_first_arg = TRUE;
					
				p = write_expr_to_str (p, ep->left, 0L, 0L);
									/* an actual arg */
			}
			*p++ = ')';
			*p = '\0';
			break;
		}
		/*}}}*/

	case  ND_INT_LIT_CONST :
		sprintf (p, "%d", ep->val.intval);
		for ( ; *p; p++);	break;

	case  ND_TEMP_SCALAR :
		sprintf (p, "%c_V%d", intern_id_prefix (ep->type),
		                      ep->val.allocp->num);
		for ( ; *p; p++);	break;

	case  ND_TEMP_ARRAY :
		sprintf (p, "%c_TMP(K_%c%d)",
		        intern_id_prefix (ep->type),
		      ((ep->val.allocp->alloc_type == ALLOC_ARRAY) ? 'T' : 'M'),
		        ep->val.allocp->num);
		for ( ; *p; p++);	break;

	case  ND_CHAR_LIT_CONST :
	case  ND_COMPLEX_LIT_CONST :
	case  ND_DP_LIT_CONST :
	case  ND_LOGICAL_LIT_CONST :
	case  ND_INTERN_OBJECT :
	case  ND_REAL_LIT_CONST :
		strcpy (p, ep->val.strval);
		for ( ; *p; p++);	break;

	/*------------------------------*/
	/* '*ep' is an operator node	*/
	/*------------------------------*/
	default :
		/*{{{  print expression tree '*ep'*/
		{
			int32  this_op        = ep->token;
			int32  prec_this_op   = this_op   & ND_PREC_MASK;
			int32  prec_parent_op = parent_op & ND_PREC_MASK;
			bool   need_brackets  = parent_op  &&
				( (prec_this_op < prec_parent_op) ||
				  (( prec_this_op == prec_parent_op) && !(this_op & branch))
				);
		
			if (need_brackets) *p++ = '(';	
			p = write_expr_to_str (p, ep->left, this_op, ND_LEFT_BRANCH);
		
			switch (this_op) {
			case ND_DEFINED_OP :	strcpy (p, ep->val.sp->name);
							for ( ; *p; p++);	break;
			case ND_EQV_OP :	strcpy (p, ".EQV.");	p += 5;	break;
			case ND_NEQV_OP :	strcpy (p, ".NEQV.");	p += 6;	break;
			case ND_OR_OP :		strcpy (p, ".OR.");	p += 4;	break;
			case ND_AND_OP :	strcpy (p, ".AND.");	p += 5;	break;
			case ND_NOT_OP :	strcpy (p, ".NOT.");	p += 5;	break;
			case ND_REL_OP :	strcpy (p, ep->val.strval);
							for ( ; *p; p++);	break;
			case ND_CONCAT_OP :	strcpy (p, "//");	p += 2;	break;
			case ND_PLUS_OP :	*p++ = '+';			break;
			case ND_U_PLUS_OP :	*p++ = '+';			break;
			case ND_MINUS_OP :	*p++ = '-';			break;
			case ND_U_MINUS_OP :	*p++ = '-';			break;
			case ND_TIMES_OP :	*p++ = '*';			break;
			case ND_DIV_OP :	*p++ = '/';			break;
			case ND_POWER_OP :	strcpy (p, "**");	p += 2;	break;
			}
		
			p = write_expr_to_str (p, ep->right, this_op, ND_RIGHT_BRANCH);
			if (need_brackets) *p++ = ')';
			*p = '\0';
		
			break;
		}
		/*}}}*/
	}
	return  p;
}
/*}}}*/

/*{{{  int  expr_len (ep, parent_op, branch)           *** out of date!! ****/
int  expr_len (ep, parent_op, branch)
/*---------------------------------*/
	Expr  *ep;
	int32  parent_op;
	int32  branch;		/* 0, ND_LEFT_BRANCH or ND_RIGHT_BRANCH */
/*--------------------------------------------------------------*/
/*	Returns the length of the string written by		*/
/* 'write_expr_to_str (ep, ..)'.				*/
/*	This is called by 'install_ida_proc_ref()' to allocate	*/
/* a sufficient amount of storage to hold the 'written' form of	*/
/* 'ep', which is attached to the 'strval' member of an 'Idref'	*/
/* record.							*/
/*--------------------------------------------------------------*/
{
	int  len,  n;
	
	if (!ep)	/* NULL pointer, eg, to left subtree of a unary op */
		return  0;

	switch (ep->token) {
	/*--------------------------------------*/
	/* '*ep' is a 'leaf' (i.e. operand):	*/
	/*--------------------------------------*/
	case  ND_SIMPLE_SCALAR :
	case  ND_WHOLE_ARRAY :
	case  ND_PROC_ACTUAL_ARG :
		return  strlen (ep->val.sp->name);

	case  ND_ARRAY_SECT :
	case  ND_FUNC :
	case  ND_INTERN_FUNC :
	case  ND_SUBR :	/** Expect 'func' to be scalar, & all keyword args,
			    etc, to be resolved, only simple args left */
		/*{{{  name (arg list)*/
		{
			bool  not_first_arg = FALSE;
			
			len = strlen ((ep->token == ND_INTERN_FUNC) ?
			               ep->val.strval : ep->val.sp->name);
			len++;		/* for '(' */
			
			while (ep = ep->right) {	/* ep = a 'ND_EXPRLIST' node */
				if (not_first_arg)
					len++;		/* for ',' */
				else
					not_first_arg = TRUE;
					
				len += expr_len (ep->left, 0, 0);
								/* an actual arg */
			}
			len++;		/* for ')' */
			return  len;
		}
		/*}}}*/

	case  ND_INT_LIT_CONST :
		if ( ! (n = ep->val.intval))	return  1;  /* value == 0 */
		else if (n > 0)			return  ndigits (n);
		else				return  (1 + ndigits (-n));

	case  ND_TEMP_SCALAR :
		if ( ! (n = ep->val.allocp->num))
			return  4;			/* e.g. "K_V0" */
		else
			return  (4 + ndigits (n));	/* e.g. "K_Vn" */
	case  ND_TEMP_ARRAY :
		if ( ! (n = ep->val.allocp->num))
			return  11;			/* "K_TMP(K_T0)" */
		else
			return  (10 + ndigits (n));	/* "K_TMP(K_Tn)" */

	case  ND_CHAR_LIT_CONST :
	case  ND_COMPLEX_LIT_CONST :
	case  ND_DP_LIT_CONST :
	case  ND_LOGICAL_LIT_CONST :
	case  ND_INTERN_OBJECT :
	case  ND_REAL_LIT_CONST :
		return  strlen (ep->val.strval);

	/*------------------------------*/
	/* '*ep' is an operator node	*/
	/*------------------------------*/
	default :
		/*{{{  print expression tree '*ep'*/
		{
			int32  this_op        = ep->token;
			int32  prec_this_op   = this_op   & ND_PREC_MASK;
			int32  prec_parent_op = parent_op & ND_PREC_MASK;
			bool   need_brackets  = parent_op  &&
				( (prec_this_op < prec_parent_op) ||
				  (( prec_this_op == prec_parent_op) && !(this_op & branch))
				);
		
			len = need_brackets ? 1 : 0;		/* for '(' */
			len += expr_len (ep->left, this_op, ND_LEFT_BRANCH);
		
			switch (this_op) {
			case ND_DEFINED_OP :	len += strlen (ep->val.sp->name); break;
			case ND_REL_OP :	len += strlen (ep->val.strval);	break;
			case ND_EQV_OP :	len += 5;   /* ".EQV." */	break;
			case ND_NEQV_OP :	len += 6;   /* ".NEQV." */	break;
			case ND_OR_OP :		len += 4;   /* ".OR." */	break;
			case ND_AND_OP :	len += 5;   /* ".AND." */	break;
			case ND_NOT_OP :	len += 5;   /* ".NOT." */	break;
			case ND_CONCAT_OP :	len += 2;   /* "//" */		break;
			case ND_POWER_OP :	len += 2;   /* "**" */		break;
			case ND_PLUS_OP :	len++;      /* '+' */		break;
			case ND_U_PLUS_OP :	len++;      /* '+' */		break;
			case ND_MINUS_OP :	len++;      /* '-' */		break;
			case ND_U_MINUS_OP :	len++;      /* '-' */		break;
			case ND_TIMES_OP :	len++;      /* '*' */		break;
			case ND_DIV_OP :	len++;      /* '/' */		break;
			}
		
			len += expr_len (ep->right, this_op, ND_RIGHT_BRANCH);
			if (need_brackets) len++;		/* for ')' */
		
			return  len;
		}
		/*}}}*/
	}
}

/*}}}*/

/*------------------------------------------------------*/
/* Functions with variable length argument lists	*/
/* (different versions for ANSI & non-ANSI C)...	*/
/*------------------------------------------------------*/
#ifdef  __STDC__

/*{{{  ANSI C version*/
/*--------------------------------------------------------------*/
/* ANSI C versions of functions with variable length arg lists.	*/
/*--------------------------------------------------------------*/

/*** THESE 'ANSI C' VERSIONS HAVEN'T BEEN TRIED YET! ***/

/*----------------*/
/* Error messages */
/*----------------*/
/*{{{  void  errmsg (char  *format, ...)*/
void  errmsg (char  *format, ...)
/*-----------------------------*/
/*----------------------*/
/*  Print error message	*/
/*----------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_lineno ();
	write_err_str (format, &argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  void  begin_errmsg (char  *format, ...)*/
void  begin_errmsg (char  *format, ...)
/*-----------------------------------*/
/*--------------------------------------------------------------*/
/*  Print the beginning of an error message.			*/
/*  (This is identical to 'errmsg()' except that it omits the	*/
/*  final '\n'!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_lineno ();
	write_err_str (format, &argp);

	va_end (argp);
}
/*}}}*/
/*{{{  void  cont_errmsg (char  *format, ...)*/
void  cont_errmsg (char  *format, ...)
/*----------------------------------*/
/*--------------------------------------------------------------*/
/*  Continue an errmsg that was begun with 'begin_errmsg()'.	*/
/* (This is identical to 'errmsg()' except that it omits the	*/
/* initial lineno and the final '\n'!).				*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_str (format, &argp);

	va_end (argp);
}
/*}}}*/
/*{{{  void  end_errmsg (char  *format, ...)*/
void  end_errmsg (char  *format, ...)
/*---------------------------------*/
/*--------------------------------------------------------------*/
/*  End an errmsg that was begun with 'begin_errmsg()'.		*/
/* (This is identical to 'errmsg()' except that it omits the	*/
/* initial lineno!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_str (format, &argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*{{{  void  warning (char  *format, ...)*/
void  warning (char  *format, ...)
/*------------------------------*/
/*--------------------------------------------------------------*/
/* Print a warning message.					*/
/* (This is virtually the same as 'errmsg' at the moment -	*/
/* except it prefaces message with '(warning)'.  Eventually,	*/
/* 'errmsg' will stop output, whereas this will not!)		*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	bool     did_have_errors;

	if (give_warnings) {
		did_have_errors = has_errors;	/* save it's value! */

		va_start (argp, format);

		write_err_lineno ();
		fputs ("(warning) ", stderr);
		write_err_str (format, &argp);
		putc ('\n', stderr);

		va_end (argp);

		has_errors = did_have_errors;	/* restore it! */
	}
}
/*}}}*/
/*{{{  void  obsc_warning (char  *format, ...)*/
void  obsc_warning (char  *format, ...)
/*-----------------------------------*/
/*--------------------------------------------------------------*/
/*  Print a warning message about the use of an obsolescent	*/
/*  feature.							*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	bool     did_have_errors;

	if (give_warnings) {
		did_have_errors = has_errors;	/* save it's value! */

		va_start (argp, format);

		write_err_lineno ();
		fputs ("(warning--obsolescent feature) ", stderr);
		write_err_str (format, &argp);
		putc ('\n', stderr);

		va_end (argp);

		has_errors = did_have_errors;	/* restore it! */
	}
}
/*}}}*/
/*{{{  void  prefix_errmsg (char  *format, ...) -- NEW (NEEDEED?)*/
void  prefix_errmsg (char  *format, ...)
/*------------------------------------*/
/*----------------------------------------------------------------
	Print error message with a prefix in parentheses, e.g.:

line 21: (not in Subset HPF) DIM arguments must be constant in
Subset HPF

Should be invoked as:
	prefix_errmsg (prefix_str, str [,arg1, arg2...]);
----------------------------------------------------------------*/
{
	va_list  argp;
	char  *prefix;

	va_start (argp, format);

	write_err_lineno ();

	prefix = va_arg (argp, char *);
	fprintf (stderr, "(%s)  ", prefix);

	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  void  not_in_subset (char  *format, ...) -- NEW*/
void  not_in_subset (char  *format, ...)
/*------------------------------------*/
/*----------------------------------------------------------------
	Print error message for feature	which is not in Subset
HPF and is not supported, e.g.:

line 21: (not in Subset HPF and not supported) internal procedures: 'FN_A'
----------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_lineno ();
	fputs ("(not in Subset HPF and not supported)  ", stderr);
	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  void  not_supported (char  *format, ...) -- NEW*/
void  not_supported (char  *format, ...)
/*------------------------------------*/
/*----------------------------------------------------------------
	Print error message for feature	which is not supported
(though it's in Subset HPF!!), e.g.:

line 21: (not supported in this version) intrinsic function 'LBOUND'
----------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_lineno ();
	fputs ("(not supported in this version)  ", stderr);
	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*{{{  write_err_str (char  *format,  va_list *argpp)*/
write_err_str (char  *format,  va_list  *argpp)
/*-------------------------------------------*/
/*----------------------*/
/*  Print error message	*/
/*----------------------*/
{
	va_list  argp = *argpp;
	register char  *p;
	char  *sval;
	Expr  *ep;
	int    ival;

	has_errors = TRUE;

	for (p = format;  *p;  p++) {
		if (*p != '%') {
			putc (*p, stderr);
			continue;
		}
		switch (*++p) {
		/*{{{  'c'   (a single char)*/
		case 'c':
			/*------------------------------------------------*/
			/* %c:  a single character; write nothing if '\0' */
			/*------------------------------------------------*/
			if (ival = va_arg (argp, int))
				putc (ival, stderr);
			break;
		
		/*}}}*/
		/*{{{  'd'   (an integer)*/
		case 'd' :
			/*-----------------*/
			/* %d:  an integer */
			/*-----------------*/
			ival = va_arg (argp, int);
			fprintf (stderr, "%i", ival);
			break;
		
		/*}}}*/
		/*{{{  'e'   (an expression)*/
		case 'e' :
			/*----------------------------------------------------*/
			/* %e:  an expression (*not* a double as in 'printf') */
			/*----------------------------------------------------*/
			ep = va_arg (argp, Expr *);
			write_expr (stderr, ep, 0L, 0L);
			break;
		
		/*}}}*/
		/*{{{  'q' or 's'   (a string, quoted if "%q")*/
		/*--------------------------------------------------------------*/
		/* case 'q': case 's':						*/
		/* -------------------						*/
		/* Output a string (usually a name).  For "%q" the string	*/
		/* is enclosed in single quotes.				*/
		/*								*/
		/* N.B. If the char pointer 'sval' is NULL, no quotes should be	*/
		/* written and the space char immediately following the "%q"	*/
		/* or "%s", if any, should be skipped.				*/
		/*--------------------------------------------------------------*/
		case 'q':
		case 's':
			sval = va_arg (argp, char *);
			if (sval)
				/*---------------------------*/
				/* string exists - print it. */
				/*---------------------------*/
				fprintf (stderr, ((*p == 'q') ? "\'%s\'" : "%s"), sval);
		
			else if (*(p+1) == ' ')
				/*----------------------------------------------*/
				/* string is NULL - skip following blank.	*/
				/*----------------------------------------------*/
				p++;
			break;
		/*}}}*/
		/*{{{  'r'   (a quoted string; remove space & convert to upper case)*/
		/*--------------------------------------------------------------*/
		/* case 'r':							*/
		/* ---------							*/
		/* As for "%q", output a string enclosed in single quotes,	*/
		/* but remove all whitespace and convert lower case letters to	*/
		/* upper.  This is used in 'format.c' to output format items,	*/
		/* removing any embedded whitespace.				*/
		/*								*/
		/* N.B. As for "%q" & "%s", if the char pointer 'sval' is NULL,	*/
		/* no quotes should be written and the space char immediately	*/
		/* following the "%r", if any, should be skipped.		*/
		/*--------------------------------------------------------------*/
		case 'r':
			sval = va_arg (argp, char *);
			if (sval) {
				/*----------------------------------------------*/
				/* string exists - print it, removing white-	*/
				/* space and converting lower case to upper.	*/
				/*----------------------------------------------*/
				putc ('\'', stderr);
				for ( ;  ival = *sval;  sval++) {
					if (islower (ival))
						putc (toupper (ival), stderr);
		
					else if ((ival != ' ') && (ival != '\t'))
						putc (ival, stderr);
				}
				putc ('\'', stderr);
			}
			else if (*(p+1) == ' ')
				/*----------------------------------------------*/
				/* string is NULL - skip following blank.	*/
				/*----------------------------------------------*/
				p++;
		
			break;
		
		/*}}}*/
		/*{{{  'X'   (an integer in unsigned hex notation)*/
		case 'X' :
			/*------------------------------------------------------*/
			/* %X:  an integer in unsigned hex notation (upper	*/
			/* case, without leading 0X), as from 'printf ()'.	*/
			/*------------------------------------------------------*/
			ival = va_arg (argp, int);
			fprintf (stderr, "%X", ival);
			break;
		
		/*}}}*/
		/*{{{  '\0'  (isolated '%' at end of string)*/
		case '\0':
			/*------------------------------------------------------*/
			/* Isolated '%' at end of string.			*/
			/*	We must take special action, as the default	*/
			/* action (see next case) would cause the '\0' to be	*/
			/* missed!						*/
			/*------------------------------------------------------*/
			p--;		/* so '\0' handled by outer loop! */
			break;
		
		/*}}}*/
		/*{{{  default   (ignore the leading '%')*/
		default:
			/*----------------------------------*/
			/* default:  ignore the leading '%' */
			/*----------------------------------*/
			putc (*p, stderr);
			break;
		/*}}}*/
		}
	}
	*argpp = argp;
}

/*}}}*/

/*--------------------------------------------------------------------*/
/* Program output  (only for 'ADAPT'; empty unless 'V_ADAPT' defined) */
/*--------------------------------------------------------------------*/
/*{{{  void  stmt (char  *format, ...)*/
void  stmt (char  *format, ...)
/*---------------------------*/
/*------------------------------*/
/* Outputs a whole statement.	*/
/*------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	write_stmt_label ();
	write_stmt_str (format, &argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  void  begin_stmt (char  *format, ...)*/
void  begin_stmt (char  *format, ...)
/*---------------------------------*/
/*--------------------------------------------------------------*/
/*	Output the start of a new statement (including the	*/
/* label field).						*/
/* N.B. This is identical to 'stmt()', except that it omits the	*/
/* final 'putchar ('\n') and does not set outcol = 1' !		*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	write_stmt_label ();
	write_stmt_str (format, &argp);

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  void  cont_stmt (char  *format, ...)*/
void  cont_stmt (char  *format, ...)
/*--------------------------------*/
/*--------------------------------------------------------------*/
/*	Continues the output of a statement.			*/
/* (This is the same as 'stmt()' except that it does not write	*/
/* an initial label field or a final '\n').			*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	write_stmt_str (format, &argp);

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  void  end_stmt (char  *format, ...)*/
void  end_stmt (char  *format, ...)
/*-------------------------------*/
/*--------------------------------------------------------------*/
/*	Outputs the remiander of the current statement.		*/
/* (This is the same as 'stmt()' except that it does not write	*/
/* an initial label field).					*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	write_stmt_str (format, &argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}
/*}}}*/
/*{{{  void  contn (char  *format, ...)*/
void  contn (char  *format, ...)
/*----------------------------*/
/*----------------------------------------------------------------------*/
/*	Outputs a whole continuation line (one that is specifically	*/
/* requested by ADAPT, rather than one that happens automatically	*/
/* because the output goes beyond column 72).				*/
/* The continuation line starts "     * ".				*/
/*----------------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	fputs ("     * ", stdout);
	outcol = 8;
	write_stmt_str (format, &argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  void  begin_contn (char  *format, ...)*/
void  begin_contn (char  *format, ...)
/*----------------------------------*/
/*----------------------------------------------------------------------*/
/*	Output the start of a continuation line (one that is		*/
/* specifically requested by ADAPT, rather than one that happens	*/
/* automatically because the output goes beyond col 72).		*/
/* The continuation line starts "     * ".				*/
/*									*/
/* N.B. This is identical to 'contn()', except that it omits		*/
/* the final 'putchar ('\n') and doesn't set outcol = 1' !		*/
/*----------------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp, format);

	fputs ("     * ", stdout);
	outcol = 8;
	write_stmt_str (format, &argp);

	va_end (argp);
#endif
}

/*}}}*/

/*{{{  write_stmt_str (char  *format,  va_list  *argpp)*/
write_stmt_str (char  *format,  va_list  *argpp)
/*--------------------------------------------*/
/*--------------------------------------------------------------*/
/* Print a statement (or part of one) of the output program.	*/
/* The following format conversion characters are used:		*/
/*	c -- char						*/
/*	d -- integer						*/
/*	n -- (Symbol *) pointer for a user ID.  O/p the name	*/
/*	                   with a prefix "%c_" prepended.	*/
/*	N -- (Symbol *) pointer for a dummy array which will be	*/
/*	             remapped. O/p it with a *special* prefix.	*/
/*	p -- int representing a datatype flag.  O/p a prefix	*/
/*	             "%c_" for an *internally* generated name	*/
/*	             of this type.				*/
/*	s -- string						*/
/* and for expressions...					*/
/*	a -- actual arg expr					*/
/*	e -- normal F90 style expr				*/
/*	E -- elementalised expr (only used in			*/
/*	                         'write_assignment_stmt ()')	*/
/*	m -- actual arg + mapvec (if array-valued)		*/
/*								*/
/* N.B. The last 4 differ only in the output format of arrays	*/
/* (i.e. elements, sections and whole arrays, both declared and	*/
/* temporary.							*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	/*{{{  decls*/
	va_list  argp = *argpp;
	register char  *p, *sval;
	char     intbuf [32];
	int      ival;
	Expr    *ep;
	Symbol  *sp;
	/*}}}*/
	
	if (has_errors)
		return;		/* don't generate output */

	for (p = format;  *p;  p++) {
		if (*p != '%') {
			writechar (*p);
			continue;
		}
		switch (*++p) {
		/*{{{  'c'  (a single char)*/
		case 'c':	/* %c:  character; write nothing if '\0' */
		
			if (ival = va_arg (argp, int))
				writechar (ival);
			break;
		
		/*}}}*/
		/*{{{  'd'  (an integer)*/
		case 'd' :	/* %d:  integer */
		
			ival = va_arg (argp, int);
			sprintf (intbuf, "%i", ival);
			for (sval = intbuf;  *sval;  sval++)
				writechar (*sval);
			break;
		
		/*}}}*/
		/*{{{  'a', 'e', 'E', 'm'  (various forms of expression)*/
		/*----------------------------------------------------------------
		Various forms of expression:
			a -- actual arg expr;
			e -- normal F90 style expr (used in I/O and for scalar exprs);
			E -- elementalised expr (only used in 'write_assignment_stmt ()')
			m -- actual arg + mapvec (if array-valued).
		
		N.B.  '%e' and '%E' *don't* mean 'double' as in 'printf'!
		----------------------------------------------------------------*/
		case 'a' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ARG_EXPR, 0L, 0L);
			break;
		
		case 'e' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, F90_EXPR, 0L, 0L);
			break;
		
		case 'E' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ELEMENTALISED_EXPR, 0L, 0L);
			break;
		
		case 'm' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ARG_PLUS_MAPVEC_EXPR, 0L, 0L);
			break;
		/*}}}*/
		/*{{{  'n'  ((Symbol *) pointer for a user ID)*/
		case 'n':
			/*------------------------------------------------------*/
			/* %n:  a (Symbol *) pointer for a user ID.		*/
			/* Output 'sp->name' with a prefix "%c_" prepended.	*/
			/*------------------------------------------------------*/
			sp = va_arg (argp, Symbol *);
		
			if ( (sp->attrs & AT_PROCEDURE)  &&
			     ( ! (sp->attrs & AT_DUMMY)  ||
			         (sp->attrs & (AT_INFO | AT_EXTERNAL)) ) ) {
				/*-------------------------------------------------
				A function name is converted to a subroutine name
				(i.e. with an untyped prefix: "O_...") *unless*
				it's a dummy argument that was unknown to be a
				function when the decls were processed, in which
				case it was declared with a datatype prefix and
				must continue to be output thus.
				-------------------------------------------------*/
				if ( ! (sp->attrs & AT_INFO)  ||
				     ! (sp->info.sp->type & TY_ELEMENTAL) ) {
		
					writechar ('O');
					writechar ('_');
				}
				/*----------------------------------------------*/
				/* ...else it's an elemental procedure, so	*/
				/* don't prepend *any* prefix!			*/
				/*----------------------------------------------*/
			}
			else {
				switch (sp->type & TY_BASIC_TYPE) {
				case  TY_INTEGER :	writechar ('I');	break;
				case  TY_REAL :		writechar ('R');	break;
				case  TY_DP :		writechar ('D');	break;
				case  TY_COMPLEX :	writechar ('X');	break;
				case  TY_LOGICAL :	writechar ('L');	break;
				case  TY_CHARACTER :	writechar ('A');	break;
				default :  /* TY_NONE or 0 */
							writechar ('O');	break;
				}
				writechar ('_');
			}
		
			for (sval = sp->name;  *sval;  sval++)
				writechar (*sval);
			break;
		/*}}}*/
		/*{{{  'N'  (    "      "      "    a dummy arg array)*/
		case 'N':
			/*------------------------------------------------------*/
			/* %N:  a (Symbol *) pointer for a user ID which is a	*/
			/* dummy argument array that will be remapped.		*/
			/*	Output 'sp->name' with a *special* prefix "%c_"	*/
			/* prepended (where '%c' is one letter greater than for	*/
			/* the '%n' case).					*/
			/*------------------------------------------------------*/
			sp = va_arg (argp, Symbol *);
		
			switch (sp->type & TY_BASIC_TYPE) {
			case  TY_INTEGER :	writechar ('J');	break;
			case  TY_REAL :		writechar ('S');	break;
			case  TY_DP :		writechar ('E');	break;
			case  TY_COMPLEX :	writechar ('Y');	break;
			case  TY_LOGICAL :	writechar ('M');	break;
			case  TY_CHARACTER :	writechar ('B');	break;
			default :  /* TY_NONE or 0 */
						writechar ('P');	break;
			}
			writechar ('_');
		
			for (sval = sp->name;  *sval;  sval++)
				writechar (*sval);
			break;
		/*}}}*/
		/*{{{  'p'  (an int for a datatype flag.  O/p internal prefix)*/
		case 'p':
			/*------------------------------------------------------*/
			/* %p:  an int representing a datatype flag.		*/
			/*	Output a prefix "%c_" for an *internally*	*/
			/* generated name of this type (where '%c' is one	*/
			/* letter greater than for the '%N' case).		*/
			/*------------------------------------------------------*/
			ival = va_arg (argp, int);
		
			switch (ival & TY_BASIC_TYPE) {
			case  TY_INTEGER :	writechar ('K');	break;
			case  TY_REAL :		writechar ('T');	break;
			case  TY_DP :		writechar ('F');	break;
			case  TY_COMPLEX :	writechar ('Z');	break;
			case  TY_LOGICAL :	writechar ('N');	break;
			case  TY_CHARACTER :	writechar ('C');	break;
			default :  /* TY_NONE or 0 */
						writechar ('Q');	break;
			}
			writechar ('_');
			break;
		
		/*}}}*/
		/*{{{  's'  (a string)*/
		case 's':	/* %s:  string; write nothing if NULL  */
		
			for (sval = va_arg (argp, char *);  *sval;  sval++)
				writechar (*sval);
			break;
		
		/*}}}*/
		/*{{{  '\0' (isolated '%' at end of string)*/
		case '\0':	/* isolated '%' at end of string! */
		
			p--;	break;		/* handled by outer loop! */
		
		/*}}}*/
		/*{{{  default  (ignore the leading '%')*/
		default:	/* default:  ignore the leading '%' */
		
			writechar (*p);	break;
		/*}}}*/
		}
	}
#endif
}
/*}}}*/

/*-------------------------------------------------*/
/* Test messages.  (All these calls are temporary) */
/*-------------------------------------------------*/
/*{{{  void  testmsg (char  *format, ...)*/
void  testmsg (char  *format, ...)
/*------------------------------*/
/*------------------------------*/
/*  Print a test message. 	*/
/*------------------------------*/
{
	va_list  argp;
	int  i;

	va_start (argp, format);

	fputs ("** TEST **: ", stderr);
	for (i = 1;  i <= test_indent;  i++)
		putc (' ', stderr);

	write_err_str (format, &argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  void  begin_testmsg (char  *format, ...)*/
void  begin_testmsg (char  *format, ...)
/*------------------------------------*/
/*--------------------------------------------------------------*/
/*  Print the beginning of a test message.			*/
/*  (This is identical to 'testmsg ()' except that it omits the	*/
/*  final '\n'!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	int  i;

	va_start (argp, format);

	fputs ("** TEST **: ", stderr);
	for (i = 1;  i <= test_indent;  i++)
		putc (' ', stderr);

	write_err_str (format, &argp);

	va_end (argp);
}
/*}}}*/
/*{{{  void  cont_testmsg (char  *format, ...)*/
void  cont_testmsg (char  *format, ...)
/*-----------------------------------*/
/*--------------------------------------------------------------*/
/* Continue a test message that was begun with			*/
/* 'begin_testmsg ()'.						*/
/* (This is identical to 'testmsg ()' except that it omits the	*/
/* initial lineno and the final '\n'!).				*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_str (format, &argp);

	va_end (argp);
}
/*}}}*/
/*{{{  void  end_testmsg (char  *format, ...)*/
void  end_testmsg (char  *format, ...)
/*----------------------------------*/
/*--------------------------------------------------------------*/
/* End a test message that was begun with 'begin_testmsg ()'.	*/
/* (This is identical to 'testmsg ()' except that it omits the	*/
/* initial lineno!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp, format);

	write_err_str (format, &argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*-----------------------------------------------------------------*/
/* other miscellaneous fns that use variable length argument lists */
/*-----------------------------------------------------------------*/
/*{{{  Expr  *intern_func_ref (char  *func_name,  int  nargs,  ...)*/
Expr  *intern_func_ref (char  *func_name,  int  nargs,  ...)
/*--------------------------------------------------------*/
	va_dcl
/*----------------------------------------------------------------
	The argument list is:

	(char  *func_name;  int N;  Expr *arg1, ... argN)

This function builds an expression tree for a reference to a
function called 'func_name'.  'N' the number of actual arguments
(>= 0), which are given by expressions 'arg1, ..., argN' (*not*
yet linked to 'ND_EXPRLIST' expression nodes!).
	This is used to build internally-generated function
references (usually to integer intrinsic functions such as "MAX",
"MIN", "MOD", etc).  They're distinguished by having (token ==
'ND_INTERN_FUNC'), and point to a string rather than an 'sp' for
the function name, since it may not have a symbol table entry.
We assume that the reference has type (EX_INTEGER | EX_SCALAR).
----------------------------------------------------------------*/
{
	va_list  argp;
	int      i;
	Expr    *arg_ep,  *func_ep,  *tail_ep;


	/*------------------------------------------------------*/
	/* create and initialise the 'ND_INTERN_FUNC' node	*/
	/*------------------------------------------------------*/
	func_ep = (Expr *) ecalloc (sizeof (Expr));	/* zeroed! */
	func_ep->token      = ND_INTERN_FUNC;
	func_ep->val.strval = func_name;		/* a static string! */
	func_ep->type       = (EX_INTEGER | EX_SCALAR);

	/*------------------------------------------------------*/
	/* link the args together in an 'arg-list', and attach	*/
	/* it to 'func_ep->right'				*/
	/*------------------------------------------------------*/
	va_start (argp, nargs);

	for (i = 0,  tail_ep = func_ep;   i < nargs;   i++) {
		arg_ep = va_arg (argp, Expr *);
		arg_ep = arg (arg_ep);
		
		attach_expr (tail_ep->right, arg_ep);
		tail_ep = arg_ep;
	}
	va_end (argp);

	return   func_ep;
}
/*}}}*/

/*}}}*/

#else

/*{{{  'K & R' C version*/
/*-----------------------------------------------------------------*/
/* 'K & R' C versions of functions with variable length arg lists. */
/*-----------------------------------------------------------------*/
/*----------------*/
/* Error messages */
/*----------------*/
/*{{{  errmsg (va_alist)*/
errmsg (va_alist)
/*--------------*/
	va_dcl
/*----------------------*/
/*  Print error message	*/
/*----------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_lineno ();
	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  begin_errmsg (va_alist)*/
begin_errmsg (va_alist)
/*-------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*  Print the beginning of an error message.			*/
/*  (This is identical to 'errmsg()' except that it omits the	*/
/*  final '\n'!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_lineno ();
	write_err_str (&argp);

	va_end (argp);
}
/*}}}*/
/*{{{  cont_errmsg (va_alist)*/
cont_errmsg (va_alist)
/*------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*  Continue an errmsg that was begun with 'begin_errmsg()'.	*/
/* (This is identical to 'errmsg()' except that it omits the	*/
/* initial lineno and the final '\n'!).				*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_str (&argp);

	va_end (argp);
}
/*}}}*/
/*{{{  end_errmsg (va_alist)*/
end_errmsg (va_alist)
/*-----------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*  End an errmsg that was begun with 'begin_errmsg()'.		*/
/* (This is identical to 'errmsg()' except that it omits the	*/
/* initial lineno!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*{{{  warning (va_alist)*/
warning (va_alist)
/*---------------*/
	va_dcl
/*--------------------------------------------------------------*/
/* Print a warning message.					*/
/* (This is virtually the same as 'errmsg' at the moment -	*/
/* except it prefaces message with '(warning)'.  Eventually,	*/
/* 'errmsg' will stop output, whereas this will not!)		*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	bool     did_have_errors;

	if (give_warnings) {
		did_have_errors = has_errors;	/* save it's value! */

		va_start (argp);

		write_err_lineno ();
		fputs ("(warning) ", stderr);
		write_err_str (&argp);
		putc ('\n', stderr);

		va_end (argp);

		has_errors = did_have_errors;	/* restore it! */
	}
}
/*}}}*/
/*{{{  obsc_warning (va_alist)*/
obsc_warning (va_alist)
/*-------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*  Print a warning message about the use of an obsolescent	*/
/*  feature.							*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	bool     did_have_errors;

	if (give_warnings) {
		did_have_errors = has_errors;	/* save it's value! */

		va_start (argp);

		write_err_lineno ();
		fputs ("(warning--obsolescent feature) ", stderr);
		write_err_str (&argp);
		putc ('\n', stderr);

		va_end (argp);

		has_errors = did_have_errors;	/* restore it! */
	}
}
/*}}}*/
/*{{{  prefix_errmsg (va_alist) -- NEW (NEEDED?)*/
prefix_errmsg (va_alist)
/*--------------------*/
	va_dcl
/*----------------------------------------------------------------
	Print error message with a prefix in parentheses, e.g.:

line 21: (not in Subset HPF) DIM arguments must be constant in
Subset HPF

Should be invoked as:
	prefix_errmsg (prefix_str, str [,arg1, arg2...]);
----------------------------------------------------------------*/
{
	va_list  argp;
	char  *prefix;

	va_start (argp);

	write_err_lineno ();

	prefix = va_arg (argp, char *);
	fprintf (stderr, "(%s)  ", prefix);

	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  not_in_subset (va_alist) -- NEW*/
not_in_subset (va_alist)
/*--------------------*/
	va_dcl
/*----------------------------------------------------------------
	Print error message for feature	which is not in Subset
HPF and is not supported, e.g.:

line 21: (not in Subset HPF and not supported) internal procedures: 'FN_A'
----------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_lineno ();
	fputs ("(not in Subset HPF and not supported)  ", stderr);
	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  not_supported (va_alist) -- NEW*/
not_supported (va_alist)
/*--------------------*/
	va_dcl
/*----------------------------------------------------------------
	Print error message for feature	which is not supported
(though it's in Subset HPF!!), e.g.:

line 21: (not supported in this version) intrinsic function 'LBOUND'
----------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_lineno ();
	fputs ("(not supported in this version)  ", stderr);
	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*{{{  write_err_str (argpp) -- OK*/
write_err_str (argpp)
/*-----------------*/
	va_list  *argpp;
/*----------------------*/
/*  Print error message	*/
/*----------------------*/
{
	va_list  argp = *argpp;
	register char  *p;
	char  *sval;
	Expr  *ep;
	int    ival;

	has_errors = TRUE;

	for (p = va_arg (argp, char *);  *p;  p++) {
		if (*p != '%') {
			putc (*p, stderr);
			continue;
		}
		switch (*++p) {
		/*{{{  'c'   (a single char)*/
		case 'c':
			/*------------------------------------------------*/
			/* %c:  a single character; write nothing if '\0' */
			/*------------------------------------------------*/
			if (ival = va_arg (argp, int))
				putc (ival, stderr);
			break;
		
		/*}}}*/
		/*{{{  'd'   (an integer)*/
		case 'd' :
			/*-----------------*/
			/* %d:  an integer */
			/*-----------------*/
			ival = va_arg (argp, int);
			fprintf (stderr, "%i", ival);
			break;
		
		/*}}}*/
		/*{{{  'e'   (an expression)*/
		case 'e' :
			/*----------------------------------------------------*/
			/* %e:  an expression (*not* a double as in 'printf') */
			/*----------------------------------------------------*/
			ep = va_arg (argp, Expr *);
			write_expr (stderr, ep, 0L, 0L);
			break;
		
		/*}}}*/
		/*{{{  'q' or 's'   (a string, quoted if '%q')*/
		/*--------------------------------------------------------------*/
		/* case 'q': case 's':						*/
		/* -------------------						*/
		/* Output a string (usually a name).  For "%q" the string	*/
		/* is enclosed in single quotes.				*/
		/*								*/
		/* N.B. If the char pointer 'sval' is NULL, no quotes should be	*/
		/* written and the space char immediately following the "%q"	*/
		/* or "%s", if any, should be skipped.				*/
		/*--------------------------------------------------------------*/
		case 'q':
		case 's':
			sval = va_arg (argp, char *);
			if (sval)
				/*---------------------------*/
				/* string exists - print it. */
				/*---------------------------*/
				fprintf (stderr, ((*p == 'q') ? "\'%s\'" : "%s"), sval);
		
			else if (*(p+1) == ' ')
				/*----------------------------------------------*/
				/* string is NULL - skip following blank.	*/
				/*----------------------------------------------*/
				p++;
		
			break;
		/*}}}*/
		/*{{{  'r'   (a quoted string; remove space & convert to upper case)*/
		/*--------------------------------------------------------------*/
		/* case 'r':							*/
		/* ---------							*/
		/* As for "%q", output a string enclosed in single quotes,	*/
		/* but remove all whitespace and convert lower case letters to	*/
		/* upper.  This is used in 'format.c' to output format items,	*/
		/* removing any embedded whitespace.				*/
		/*								*/
		/* N.B. As for "%q" & "%s", if the char pointer 'sval' is NULL,	*/
		/* no quotes should be written and the space char immediately	*/
		/* following the "%r", if any, should be skipped.		*/
		/*--------------------------------------------------------------*/
		case 'r':
			sval = va_arg (argp, char *);
			if (sval) {
				/*----------------------------------------------*/
				/* string exists - print it, removing white-	*/
				/* space and converting lower case to upper.	*/
				/*----------------------------------------------*/
				putc ('\'', stderr);
				for ( ;  ival = *sval;  sval++) {
					if (islower (ival))
						putc (toupper (ival), stderr);
		
					else if ((ival != ' ') && (ival != '\t'))
						putc (ival, stderr);
				}
				putc ('\'', stderr);
			}
			else if (*(p+1) == ' ')
				/*----------------------------------------------*/
				/* string is NULL - skip following blank.	*/
				/*----------------------------------------------*/
				p++;
		
			break;
		/*}}}*/
		/*{{{  'X'   (an integer in unsigned hex notation)*/
		case 'X' :
			/*------------------------------------------------------*/
			/* %X:  an integer in unsigned hex notation (upper	*/
			/* case, without leading 0X), as from 'printf ()'.	*/
			/*------------------------------------------------------*/
			ival = va_arg (argp, int);
			fprintf (stderr, "%X", ival);
			break;
		/*}}}*/
		/*{{{  '\0'  (isolated '%' at end of string)*/
		case '\0':
			/*------------------------------------------------------*/
			/* Isolated '%' at end of string.			*/
			/*	We must take special action, as the default	*/
			/* action (see next case) would cause the '\0' to be	*/
			/* missed!						*/
			/*------------------------------------------------------*/
			p--;		/* so '\0' handled by outer loop! */
			break;
		
		/*}}}*/
		/*{{{  default   (ignore the leading '%')*/
		default:
			/*----------------------------------*/
			/* default:  ignore the leading '%' */
			/*----------------------------------*/
			putc (*p, stderr);
			break;
		/*}}}*/
		}
	}
	*argpp = argp;
}

/*}}}*/

/*--------------------------------------------------------------------*/
/* Program output  (only for 'ADAPT'; empty unless 'V_ADAPT' defined) */
/*--------------------------------------------------------------------*/
/*{{{  stmt (va_alist)*/
stmt (va_alist)
/*-----------*/
	va_dcl
/*------------------------------*/
/* Outputs a whole statement.	*/
/*------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	write_stmt_label ();
	write_stmt_str (&argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  begin_stmt (va_alist)*/
begin_stmt (va_alist)
/*-----------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*	Output the start of a new statement (including the	*/
/* label field).						*/
/* N.B. This is identical to 'stmt()', except that it omits the	*/
/* final 'putchar ('\n') and does not set outcol = 1' !		*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	write_stmt_label ();
	write_stmt_str (&argp);

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  cont_stmt (va_alist)*/
cont_stmt (va_alist)
/*----------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*	Continues the output of a statement.			*/
/* (This is the same as 'stmt()' except that it does not write	*/
/* an initial label field or a final '\n').			*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	write_stmt_str (&argp);

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  end_stmt (va_alist)*/
end_stmt (va_alist)
/*---------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*	Outputs the remiander of the current statement.		*/
/* (This is the same as 'stmt()' except that it does not write	*/
/* an initial label field).					*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	write_stmt_str (&argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  contn (va_alist)*/
contn (va_alist)
/*------------*/
	va_dcl
/*----------------------------------------------------------------------*/
/*	Outputs a whole continuation line (one that is specifically	*/
/* requested by ADAPT, rather than one that happens automatically	*/
/* because the output goes beyond column 72).				*/
/* The continuation line starts "     * ".				*/
/*----------------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	fputs ("     * ", stdout);
	outcol = 8;
	write_stmt_str (&argp);
	putchar ('\n');
	outcol = 1;

	va_end (argp);
#endif
}

/*}}}*/
/*{{{  begin_contn (va_alist)*/
begin_contn (va_alist)
/*------------------*/
	va_dcl
/*----------------------------------------------------------------------*/
/*	Output the start of a continuation line (one that is		*/
/* specifically requested by ADAPT, rather than one that happens	*/
/* automatically because the output goes beyond col 72).		*/
/* The continuation line starts "     * ".				*/
/*									*/
/* N.B. This is identical to 'contn()', except that it omits		*/
/* the final 'putchar ('\n') and doesn't set outcol = 1' !		*/
/*----------------------------------------------------------------------*/
{
#ifdef V_ADAPT
	va_list  argp;

	va_start (argp);

	fputs ("     * ", stdout);
	outcol = 8;
	write_stmt_str (&argp);

	va_end (argp);
#endif
}

/*}}}*/

/*{{{  write_stmt_str (argpp) -- OK*/
write_stmt_str (argpp)
/*------------------*/
	va_list  *argpp;
/*--------------------------------------------------------------*/
/* Print a statement (or part of one) of the output program.	*/
/* The following format conversion characters are used:		*/
/*	c -- char						*/
/*	d -- integer						*/
/*	n -- (Symbol *) pointer to a user ID.  O/p the name	*/
/*	                   with a prefix "%c_" prepended.	*/
/*	N -- (Symbol *) pointer to a dummy array which will be	*/
/*	             remapped. O/p it with a *special* prefix.	*/
/*	p -- int representing a datatype flag.  O/p a prefix	*/
/*	             "%c_" for an *internally* generated name	*/
/*	             of this type.				*/
/*	s -- string						*/
/* and for expressions...					*/
/*	a -- actual arg expr					*/
/*	e -- normal F90 style expr				*/
/*	E -- elementalised expr (only used in			*/
/*	                         'write_assignment_stmt ()')	*/
/*	m -- actual arg + mapvec (if array-valued)		*/
/*								*/
/* N.B. The last 4 differ only in the output format of arrays	*/
/* (i.e. elements, sections and whole arrays, both declared and	*/
/* temporary.							*/
/*--------------------------------------------------------------*/
{
#ifdef V_ADAPT
	/*{{{  decls*/
	va_list  argp = *argpp;
	register char  *p, *sval;
	char     intbuf [32];
	int      ival;
	Expr    *ep;
	Symbol  *sp;
	/*}}}*/

	if (has_errors)
		return;		/* don't generate output */

	for (p = va_arg (argp, char *);  *p;  p++) {
		if (*p != '%') {
			writechar (*p);
			continue;
		}
		switch (*++p) {
		/*{{{  'c'  (a single char)*/
		case 'c':	/* %c:  character; write nothing if '\0' */
		
			if (ival = va_arg (argp, int))
				writechar (ival);
			break;
		
		/*}}}*/
		/*{{{  'd'  (an integer)*/
		case 'd' :	/* %d:  integer */
		
			ival = va_arg (argp, int);
			sprintf (intbuf, "%i", ival);
			for (sval = intbuf;  *sval;  sval++)
				writechar (*sval);
			break;
		
		/*}}}*/
		/*{{{  'a', 'e', 'E', 'm'  (various forms of expression)*/
		/*----------------------------------------------------------------
		Various forms of expression:
			a -- actual arg expr;
			e -- normal F90 style expr (used in I/O and for scalar exprs);
			E -- elementalised expr (only used in 'write_assignment_stmt ()')
			m -- actual arg + mapvec (if array-valued).
		
		N.B.  '%e' and '%E' *don't* mean 'double' as in 'printf'!
		----------------------------------------------------------------*/
		case 'a' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ARG_EXPR, 0L, 0L);
			break;
		
		case 'e' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, F90_EXPR, 0L, 0L);
			break;
		
		case 'E' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ELEMENTALISED_EXPR, 0L, 0L);
			break;
		
		case 'm' :
			ep = va_arg (argp, Expr *);
			stmt_write_expr (ep, ARG_PLUS_MAPVEC_EXPR, 0L, 0L);
			break;
		/*}}}*/
		/*{{{  'n'  ((Symbol *) pointer for a user ID)*/
		case 'n':
			/*------------------------------------------------------*/
			/* %n:  a (Symbol *) pointer for a user ID.		*/
			/* Output 'sp->name' with a prefix "%c_" prepended.	*/
			/*------------------------------------------------------*/
			sp = va_arg (argp, Symbol *);
		
			if ( (sp->attrs & AT_PROCEDURE)  &&
			     ( ! (sp->attrs & AT_DUMMY)  ||
			         (sp->attrs & (AT_INFO | AT_EXTERNAL)) ) ) {
				/*-------------------------------------------------
				A function name is converted to a subroutine name
				(i.e. with an untyped prefix: "O_...") *unless*
				it's a dummy argument that was unknown to be a
				function when the decls were processed, in which
				case it was declared with a datatype prefix and
				must continue to be output thus.
				-------------------------------------------------*/
				if ( ! (sp->attrs & AT_INFO)  ||
				     ! (sp->info.sp->type & TY_ELEMENTAL) ) {
		
					writechar ('O');
					writechar ('_');
				}
				/*----------------------------------------------*/
				/* ...else it's an elemental procedure, so	*/
				/* don't prepend *any* prefix!			*/
				/*----------------------------------------------*/
			}
			else {
				switch (sp->type & TY_BASIC_TYPE) {
				case  TY_INTEGER :	writechar ('I');	break;
				case  TY_REAL :		writechar ('R');	break;
				case  TY_DP :		writechar ('D');	break;
				case  TY_COMPLEX :	writechar ('X');	break;
				case  TY_LOGICAL :	writechar ('L');	break;
				case  TY_CHARACTER :	writechar ('A');	break;
				default :  /* TY_NONE or 0 */
							writechar ('O');	break;
				}
				writechar ('_');
			}
		
			for (sval = sp->name;  *sval;  sval++)
				writechar (*sval);
			break;
		/*}}}*/
		/*{{{  'N'  (    "       "      "   a dummy arg array)*/
		case 'N':
			/*------------------------------------------------------*/
			/* %N:  a (Symbol *) pointer for a user ID which is a	*/
			/* dummy argument array that will be remapped.		*/
			/*	Output 'sp->name' with a *special* prefix "%c_"	*/
			/* prepended (where '%c' is one letter greater than for	*/
			/* the '%n' case).					*/
			/*------------------------------------------------------*/
			sp = va_arg (argp, Symbol *);
		
			switch (sp->type & TY_BASIC_TYPE) {
			case  TY_INTEGER :	writechar ('J');	break;
			case  TY_REAL :		writechar ('S');	break;
			case  TY_DP :		writechar ('E');	break;
			case  TY_COMPLEX :	writechar ('Y');	break;
			case  TY_LOGICAL :	writechar ('M');	break;
			case  TY_CHARACTER :	writechar ('B');	break;
			default :  /* TY_NONE or 0 */
						writechar ('P');	break;
			}
			writechar ('_');
		
			for (sval = sp->name;  *sval;  sval++)
				writechar (*sval);
			break;
		/*}}}*/
		/*{{{  'p'  (an int for a datatype flag.  O/p internal prefix)*/
		case 'p':
			/*------------------------------------------------------*/
			/* %p:  an int representing a datatype flag.		*/
			/*	Output a prefix "%c_" for an *internally*	*/
			/* generated name of this type (where '%c' is one	*/
			/* letter greater than for the '%N' case).		*/
			/*------------------------------------------------------*/
			ival = va_arg (argp, int);
		
			switch (ival & TY_BASIC_TYPE) {
			case  TY_INTEGER :	writechar ('K');	break;
			case  TY_REAL :		writechar ('T');	break;
			case  TY_DP :		writechar ('F');	break;
			case  TY_COMPLEX :	writechar ('Z');	break;
			case  TY_LOGICAL :	writechar ('N');	break;
			case  TY_CHARACTER :	writechar ('C');	break;
			default :  /* TY_NONE or 0 */
						writechar ('Q');	break;
			}
			writechar ('_');
			break;
		/*}}}*/
		/*{{{  's'  (a string)*/
		case 's':	/* %s:  string; write nothing if NULL  */
		
			for (sval = va_arg (argp, char *);  *sval;  sval++)
				writechar (*sval);
			break;
		
		/*}}}*/
		/*{{{  '\0' (isolated '%' at end of string)*/
		case '\0':	/* isolated '%' at end of string! */
		
			p--;	break;		/* handled by outer loop! */
		
		/*}}}*/
		/*{{{  default  (ignore the leading '%')*/
		default:	/* default:  ignore the leading '%' */
		
			writechar (*p);	break;
		/*}}}*/
		}
	}
#endif
}


/*}}}*/


/*-------------------------------------------------*/
/* Test messages.  (All these calls are temporary) */
/*-------------------------------------------------*/
/*{{{  testmsg (va_alist)*/
testmsg (va_alist)
/*--------------*/
	va_dcl
/*------------------------------*/
/*  Print a test message. 	*/
/*------------------------------*/
{
	va_list  argp;
	int  i;

	va_start (argp);

	fputs ("** TEST **: ", stderr);
	for (i = 1;  i <= test_indent;  i++)
		putc (' ', stderr);

	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/
/*{{{  begin_testmsg (va_alist)*/
begin_testmsg (va_alist)
/*--------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/*  Print the beginning of a test message.			*/
/*  (This is identical to 'testmsg ()' except that it omits the	*/
/*  final '\n'!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;
	int  i;

	va_start (argp);

	fputs ("** TEST **: ", stderr);
	for (i = 1;  i <= test_indent;  i++)
		putc (' ', stderr);

	write_err_str (&argp);

	va_end (argp);
}
/*}}}*/
/*{{{  cont_testmsg (va_alist)*/
cont_testmsg (va_alist)
/*-------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/* Continue a test message that was begun with			*/
/* 'begin_testmsg ()'.						*/
/* (This is identical to 'testmsg ()' except that it omits the	*/
/* initial lineno and the final '\n'!).				*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_str (&argp);

	va_end (argp);
}
/*}}}*/
/*{{{  end_testmsg (va_alist)*/
end_testmsg (va_alist)
/*------------------*/
	va_dcl
/*--------------------------------------------------------------*/
/* End a test message that was begun with 'begin_testmsg ()'.	*/
/* (This is identical to 'testmsg ()' except that it omits the	*/
/* initial lineno!).						*/
/*--------------------------------------------------------------*/
{
	va_list  argp;

	va_start (argp);

	write_err_str (&argp);
	putc ('\n', stderr);

	va_end (argp);
}
/*}}}*/

/*--------------------------------------*/
/* other misc. fns that use "varargs.h" */
/*--------------------------------------*/
/*{{{  Expr  *intern_func_ref (va_alist) -- ok*/
Expr  *intern_func_ref (va_alist)
/*-----------------------------*/
	va_dcl
/*----------------------------------------------------------------
	The argument list is:

	(char  *func_name;  int N;  Expr *arg1, ... argN)

This function builds an expression tree for a reference to a
function called 'func_name'.  'N' the number of actual arguments
(>= 0), which are given by expressions 'arg1, ..., argN' (*not*
yet linked to 'ND_EXPRLIST' expression nodes!).
	This is used to build internally-generated function
references (usually to integer intrinsic functions such as "MAX",
"MIN", "MOD", etc).  They're distinguished by having (token ==
'ND_INTERN_FUNC'), and point to a string rather than an 'sp' for
the function name, since it may not have a symbol table entry.
We assume that the reference has type (EX_INTEGER | EX_SCALAR).
----------------------------------------------------------------*/
{
	va_list  argp;
	char    *func_name;
	int      i,  nargs;
	Expr    *arg_ep,  *func_ep,  *tail_ep;

	va_start (argp);

	/*------------------------------------------------------*/
	/* create and initialise the 'ND_INTERN_FUNC' node	*/
	/*------------------------------------------------------*/
	func_name = va_arg (argp, char *);

	func_ep = (Expr *) ecalloc (sizeof (Expr));	/* zeroed! */
	func_ep->token      = ND_INTERN_FUNC;
	func_ep->val.strval = func_name;		/* a static string! */
	func_ep->type       = (EX_INTEGER | EX_SCALAR);

	/*------------------------------------------------------*/
	/* link the args together in an 'arg-list', and attach	*/
	/* it to 'func_ep->right'				*/
	/*------------------------------------------------------*/
	nargs = va_arg (argp, int);
	for (i = 0,  tail_ep = func_ep;   i < nargs;   i++) {
		arg_ep = va_arg (argp, Expr *);
		arg_ep = exprlist_item (arg_ep);
		
		attach_expr (tail_ep->right, arg_ep);
		tail_ep = arg_ep;
	}
	va_end (argp);

	return   func_ep;
}
/*}}}*/
/*}}}*/

#endif
/*}}}*/
	
